-- Ouphoria v0.1 pre-alpha preprocessor support library
-- Created by Jrgen Jnes

without warning
-------------------------------------------------------------------------------------------------
-- VARIABLES
-------------------------------------------------------------------------------------------------
sequence class_names, class_data

sequence data
data = {}

-------------------------------------------------------------------------------------------------
-- ERROR HANDLING
-------------------------------------------------------------------------------------------------
procedure AbortError (sequence error_message)
integer Handle

Handle = open ("ox.err", "w")
if equal (Handle, -1) then
	puts (1, "can't open ox.err\n")
	abort (0)
end if
puts (Handle, error_message)
close (Handle)
puts (1, error_message)
abort (0)
end procedure
-------------------------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------------------------
constant
    True    = 1,
    False   = 0

-------------------------------------------------------------------------------------------------
type value (object x)
return True
end type
-------------------------------------------------------------------------------------------------
global type object (value x)
	return True
end type
-------------------------------------------------------------------------------------------------
global type atom (object x)
if equal (x, 0) then
	return True
end if
if equal (x / x, 1) then
	return True
end if
return False
end type
-------------------------------------------------------------------------------------------------
global type integer (object x)
if equal (x, 0) then
	return True
end if
if equal (x / floor (x), 1) then
	return True
end if
return False
end type
-------------------------------------------------------------------------------------------------
global type sequence (object x)
if not atom (x) then
	return True
end if
return False
end type
-------------------------------------------------------------------------------------------------
type string (object x)
if not sequence (x) then 
	return False 
end if
for Index = 1 to length (x) do
   if not integer (x[Index]) or
       x[Index] < 0 then
        return False
    end if
end for
return True
end type
-------------------------------------------------------------------------------------------------
type subscripts (sequence x)
for Index = 1 to length (x) do
	if not integer (x[Index]) or
	   x[Index] < 0 then
		return False
	end if
end for
return True
end type
-------------------------------------------------------------------------------------------------
type path (sequence x)
for Index = 1 to length (x) do
	if not integer (x[Index]) and
	   not string  (x[Index]) then
		return False
	end if
end for
return True
end type
-------------------------------------------------------------------------------------------------
global type pointer (integer x)
if not integer (x) then
	return False
end if
if length (data) < x or
              x < 1 then
	return False
end if
return True
end type
-------------------------------------------------------------------------------------------------
global type collection (sequence x)
if not string (x) then
	return False
end if
for Index = 1 to length (x) do
	if not object (x[Index]) then
		return False
	end if
end for
return True
end type
-------------------------------------------------------------------------------------------------
-- ROUTINES FOR MANIPULATING TEMPLATE CLASSES
-------------------------------------------------------------------------------------------------
global procedure class (string Name, string Parent)
integer Parent_Index

-- find the right template/abort when not found
Parent_Index = find (Parent, class_names)
if Parent_Index = 0 then
AbortError ("template class " & Parent & " has not been declared")
end if

-- create the actual template
class_names = append (class_names, Name)
class_data  = append (class_data , class_data[Parent_Index])
end procedure
-------------------------------------------------------------------------------------------------
global procedure end_class () -- this is reserved for future use
end procedure
-------------------------------------------------------------------------------------------------
constant
	NAME     = 1,
	TYPE     = 2,
	ONREAD   = 3,
	ONWRITE  = 4,
	VALUE    = 5,
	HASVALUE = 6 
-------------------------------------------------------------------------------------------------
global procedure property (string Name, integer Type, string onRead, string onWrite)
class_data[length (class_data)] = 
append (class_data[length (class_data)], {Name, Type, class_names[length (class_names)] & '_' & onRead, class_names[length (class_names)] & '_' & onWrite, 0, False} )
end procedure
-------------------------------------------------------------------------------------------------
constant
	ID = 2
-------------------------------------------------------------------------------------------------
global procedure method (string Name)
class_data[length (class_data)] = append (class_data[length (class_data)], {Name, class_names[length (class_names)] & "_" & Name})
end procedure
-------------------------------------------------------------------------------------------------
global procedure assign_method (string method_name, integer Id)
for Index = length (class_data) to 1 by -1 do
	for Index2 = length (class_data[Index]) to 1 by -1 do
		if equal (class_data[Index][Index2][ID], method_name) then
			class_data[Index][Index2][ID] = Id
		end if
	end for
end for
end procedure
-------------------------------------------------------------------------------------------------
-- ROUTINES FOR MANIPULATING REAL CLASSES
-------------------------------------------------------------------------------------------------
	integer READ_PROPERTY
	procedure SET_READ_PROPERTY (integer x)
	READ_PROPERTY = x
	end procedure
-------------------------------------------------------------------------------------------------
global procedure call_pmethod (path Path, object Parameters)
integer Sucess
Sucess = False
if length (Path) > 2 then
	Path = {call_func (READ_PROPERTY, Path[1..length (Path - 1)]) , length (Path)}
end if
for Index = length (data[Path[1]]) to 1 by -1 do
	if equal (data[Path[1]][Index][NAME], Path[2]) then
		call_proc (data[Path[1]][Index][ID], Path[1] & Parameters)
		Sucess = True
	end if
end for
if Sucess = False then
	AbortError ("method " & Path[2] & " has not been declared")
end if
if equal (Path[2], "Destroy") then
	data[Path[1]] = 0
end if 
end procedure
-------------------------------------------------------------------------------------------------
global function call_fmethod (path Path, object Parameters)
if length (Path) > 2 then
	Path = {call_func (routine_id (READ_PROPERTY), Path[1..length (Path - 1)]) , length (Path)}
end if
for Index = length (data[Path[1]]) to 1 by -1 do
	if equal (data[Path[1]][Index][NAME], Path[2]) then
		return call_func (data[Path[1]][Index][ID], Path[1] & Parameters)
	end if
end for
AbortError ("method " & Path[2] & " has not been declared")
end function
-------------------------------------------------------------------------------------------------
	function readFromSeq (sequence Sequence, subscripts Subscripts)
	if equal (Subscripts, {}) then
		return Sequence
	end if
	for Index = 1 to length (Subscripts) do
    	Sequence = Sequence[Subscripts[Index]]
	end for
	return Sequence
	end function
-------------------------------------------------------------------------------------------------
global function read_property (path Path)
for Index = length (data[Path[1]]) to 1 by -1 do
	if equal (data[Path[1]][Index][NAME], Path[2]) then
		if equal (data[Path[1]][Index][TYPE], routine_id ("pointer")) and
		   length (Path) > 2 then 
			if data[Path[1]][Index][HASVALUE] = False then 
				AbortError ("property " & Path[2] & " has not been assigned a value")
			end if
			if not string (data[Path[1]][Index][ONREAD]) then 
				call_pmethod ({Path[1], data[Path[1]][Index][ONREAD]}, {Path[1]})
			end if 
			Path[2] = data[Path[1]][Index][VALUE]
			Path = Path[2..length (Path)]
			return read_property (Path)

		elsif equal (data[Path[1]][Index][TYPE], routine_id ("collection")) and
		      length (Path) > 2 then 
			if data[Path[1]][Index][HASVALUE] = False then 
				AbortError ("property " & Path[2] & " has not been assigned a value")
			end if
			if not string (data[Path[1]][Index][ONREAD]) then 
				call_pmethod ({Path[1], data[Path[1]][Index][ONREAD]}, {Path[1]})
			end if 
			Path[3] = data[Path[1]][Index][VALUE][Path[3]]
			Path = Path[3..length (Path)]
			return read_property (Path)

		else -- if the property is of type "normal"
			if data[Path[1]][Index][HASVALUE] = False then 
				AbortError ("property " & Path[2] & " has not been assigned a value")
			end if
			if not string (data[Path[1]][Index][ONREAD]) then
				call_pmethod ({Path[1], Path[2]}, {Path[1]})
			end if 
			if length (Path) > 2 then
				return readFromSeq (data[Path[1]][Index][VALUE], Path[3..length (Path)])
			end if
			return data[Path[1]][Index][VALUE]
		end if
	end if
end for
AbortError ("property " & Path[2] & " does not exist")
end function
-------------------------------------------------------------------------------------------------
SET_READ_PROPERTY (routine_id ("read_property"))
-------------------------------------------------------------------------------------------------
	function writeToSeq (sequence Sequence, subscripts Subscripts, object Value)
	sequence Temp, Temp2
	Temp = {}
	Temp2 = Sequence
	for Index = 1 to length (Subscripts) do
    	Temp = append (Temp, Temp2[ Subscripts[Index] ])
    	Temp2 = Temp2[Subscripts[Index]]
	end for
	Temp = prepend (Temp, Sequence)
	Temp[length (Temp)] = Value
	for Index = length (Temp) - 1 to 1 by -1 do
    	Temp[Index][Subscripts[Index]] = Temp[Index + 1]
	end for
	Temp = Temp[1]
	return Temp
	end function
-------------------------------------------------------------------------------------------------
global procedure write_property (path Path, object Value)
integer Sucess
Sucess = False

for Index = length (data[Path[1]]) to 1 by -1 do
	if equal (data[Path[1]][Index][NAME], Path[2]) then
		if equal (data[Path[1]][Index][TYPE], routine_id ("pointer")) and
		   length (Path) > 2 then
			if data[Path[1]][Index][HASVALUE] = False then 
				AbortError ("property " & Path[2] & " has not been assigned a value")
			end if
			if not string (data[Path[1]][Index][ONWRITE]) then 
				call_pmethod ({Path[1], data[Path[1]][Index][ONWRITE]}, {Path[1]})
			end if 
			Path[2] = data[Path[1]][Index][VALUE]
			Path = Path[2..length (Path)]
			write_property (Path, Value)
			Sucess = True
			exit

		elsif equal (data[Path[1]][Index][TYPE], routine_id ("collection")) and
		      length (Path) > 2 then 
			if data[Path[1]][Index][HASVALUE] = False then 
				AbortError ("property " & Path[2] & " has not been assigned a value")
			end if
			if not string (data[Path[1]][Index][ONWRITE]) then 
				call_pmethod ({Path[1], data[Path[1]][Index][ONWRITE]}, {Path[1]})
			end if 
			Path[3] = data[Path[1]][Index][VALUE][Path[3]]
			Path = Path[3..length (Path)]
			write_property (Path, Value)
			Sucess = True
			exit

		else -- if the property is of type "normal"
			if not string (data[Path[1]][Index][ONREAD]) then 
				call_pmethod ({Path[1], Path[2]}, {Path[1]})
			end if 
			if length (Path) > 2 then
				data[Path[1]][Index][VALUE] = writeToSeq (data[Path[1]][Index][VALUE], Path[3..length (Path)], Value)
			end if
			if call_func (data[Path[1]][Index][TYPE], {Value}) then
				data[Path[1]][Index][VALUE] = Value
				data[Path[1]][Index][HASVALUE] = True
				Sucess = True
			else
				AbortError ("type_check failure in property " & data[Path[1]][Index][NAME])
			end if
		end if
	end if
end for
if not Sucess then
	AbortError ("property " & Path[2] & " does not exist")
end if
end procedure
-------------------------------------------------------------------------------------------------
global function new (string Template, sequence Constructor_Parameters)
integer Template_Index, class_pointer

-- find the right template/abort when not found
Template_Index = find (Template, class_names)
if Template_Index = 0 then
AbortError ("class " & Template & " has not been declared")
end if

-- create the new class
data = append (data, class_data[Template_Index])

class_pointer = length (data)

write_property ({length (data), "Name"}, "class" & sprintf ("%d", {length (data)}))
write_property ({length (data), "Type"}, Template)
call_pmethod ({length (data), "Init"}, Constructor_Parameters)
return class_pointer
end function
-------------------------------------------------------------------------------------------------
-- DECLARE BASIC CLASS: TClass
-------------------------------------------------------------------------------------------------
class_names = {"TClass"} -- equals to: template ("TClass", "")
class_data  = {{}}       -- equals to: template ("TClass", "")

-- Properties:
property ("Name", routine_id ("string"), "", "")
property ("Type", routine_id ("string"), "", "")

-- Methods:
method ("Init")

-------------------------------------------------------------------------------------------------
procedure TClass_Init (pointer x)
end procedure
assign_method ("TClass_Init", routine_id ("TClass_Init"))
-------------------------------------------------------------------------------------------------
